/* blccc2bmdrv.c
 * BLINKENmini converter for blccc
 *  - receives frames on udp and displays them using bmdrv
 * Copyright (C) 2002 sphaera & 1stein (http://blinkenmini.1stein.no-ip.com/)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>

#define VerMaj 1
#define VerMin 0
#define VerRev 0

//main program
int main( int ArgC, char * * Args )
{
	//local variables
	int BmDrv, Socket, Ip, Port, AddrL, Width, Height, X, Y, I;
	char * pChr;
	struct sockaddr_in Addr;
	unsigned char Buffer[8192], BmDrvBuffer[8192], WidthB, HeightB;
	struct hostent * pHostent;

	//clear screen and print message
	printf( "\033c" );
	printf( "blccc2bmdrv: BLINKENmini converter for blccc (version %d.%d.%d)\n", VerMaj, VerMin, VerRev );
	printf( "             receives frames on udp and displays them using bmdrv\n" );
	printf( "             done in 2002 by sphaera & 1stein\n" );
	printf( "             http://blinkenmini.1stein.no-ip.com/\n" );
	printf( "             distributed under GNU Public License - no warranty\n\n" );

	//check parameters
	if( ArgC != 3 )
	{
		//diplay help message
		printf( "syntax: blccc2bmdrv <[host:]port> <out-bmdrv>\n" );
		printf( "  - [host:]port - host and port to receive data on\n" );
		printf( "  - out-bmdrv - path to output device of bmdrv\n\n" );
		return -1;
	}

	//parse ip and port
	pChr = strchr( Args[1], ':' );
	if( pChr == NULL )
	{
		Ip = INADDR_ANY;
		Port = atoi( Args[1] );
	}
	else
	{
		*pChr = 0;
		Ip = inet_addr( Args[1] );
		Port = atoi( pChr + 1 );
		//resolve hostname
		if( Ip == INADDR_NONE )
		{
			pHostent = gethostbyname( Args[1] );
			if( pHostent != NULL )
				if( pHostent->h_addr_list != NULL )
					if( pHostent->h_addr_list[0] != NULL )
						Ip = *(unsigned long *)(pHostent->h_addr_list[0]);
		}
	}	
	//no ip found
	if( Ip == INADDR_NONE )
	{
		printf( "\nblccc2bmdrv: could not resolve host\n\n" );
		return -1;
	}

	//open bmdrv
	BmDrv = open( Args[2], O_WRONLY | O_NONBLOCK );
	if( BmDrv == -1 )
	{
		printf( "\nblccc2bmdrv: could not open out-bmdrv: %s\n\n", strerror( errno ) );
		return -1;
	}

	//create socket
	Socket = socket( AF_INET, SOCK_DGRAM, 0 );
	if( Socket == -1 )
	{
		printf( "\nblccc2bmdrv: could not create socket: %s\n\n", strerror( errno ) );
		close( BmDrv );
		return -1;
	}

	//bind socket to local address
	Addr.sin_family = AF_INET;
	Addr.sin_port = htons( Port );
	Addr.sin_addr.s_addr = Ip;	
	I = bind( Socket, (struct sockaddr *)&Addr, sizeof( struct sockaddr ) );
	if( I == -1 )
	{
		printf( "\nblccc2bmdrv: could not bind socket: %s\n\n", strerror( errno ) );
		close( Socket );
		close( BmDrv );
		return -1;
	}

	//main loop
	for( ; ; )
	{
		//receive on socket
		AddrL = sizeof( Addr );
		I = recvfrom( Socket, Buffer, sizeof( Buffer ), 0, (struct sockaddr *)&Addr, &AddrL );
		if( I == -1 )
		{
			if( errno == EINTR ) //exit main loop
				break;
			printf( "\nblccc2bmdrv: error while receiving from socket: %s\n\n", strerror( errno ) );
			close( Socket );
			close( BmDrv );
			return -1;
		}
		//datagram received that was long enough
		if( I >= 12 )
		{
			//check magic
			if( Buffer[0] == 0xDE && Buffer[1] == 0xAD && Buffer[2] == 0xBE && Buffer[3] == 0xEF )
			{
				//get dimensions
				Width = (unsigned short)ntohs( *(short *)(Buffer + 8) );
				Height = (unsigned short)ntohs( *(short *)(Buffer + 10) );
				//check if enough data
				if( I >= 12 + Width * Height )
				{
					//print source address
					printf( "\033[7;1H\033[2Kframe from %d.%d.%d.%d:%d\n",
							((unsigned char *)&Addr.sin_addr.s_addr)[0], ((unsigned char *)&Addr.sin_addr.s_addr)[1],
							((unsigned char *)&Addr.sin_addr.s_addr)[2], ((unsigned char *)&Addr.sin_addr.s_addr)[3],
							ntohs( Addr.sin_port ) );
					
					//print frame to screen
					printf( "\033[9;1H" );
					for( Y = 0; Y < Height; Y++ )
					{
						for( X = 0; X < Width; X++ )
						{
							if( Buffer[12 + X + Y * Width] )
								printf( "#" );
							else
								printf( " " );
						}
						printf( "\n" );
					}
					
					//output frame to bmdrv
					if( Width <= 255 ) //reduce width to byte
						WidthB = (unsigned char)Width;
					else
						WidthB = 255;
					if( Height <= 255 ) //reduce height to byte
						HeightB = (unsigned char)Height;
					else
						HeightB = 255;
					if( 2 + Width * Height + 1 < sizeof( BmDrvBuffer ) ) //frame fits into buffer
					{
						//dimensions into buffer
						BmDrvBuffer[0] = WidthB;
						BmDrvBuffer[1] = HeightB;
						I = 2;
						//pixels into buffer
						for( Y = 0; Y < HeightB; Y++ ) //pixels
							for( X = 0; X < WidthB; X++ )
								if( Buffer[12 + X + Y * Width] )
									BmDrvBuffer[I++] = 0xFF;
								else
									BmDrvBuffer[I++] = 0x00;
						//delay into buffer (no delay)
						BmDrvBuffer[I++] = 0;
						//write buffer to bmdrv
						write( BmDrv, BmDrvBuffer, I );
					} //if( 2 + Width * Height + 1 < ...
				}
			}
		} //if( I >= 12 )
	} //for( ; ; )

	//close socket
	close( Socket );

	//close bmdrv
	close( BmDrv );

	//print end-message
	printf( "\nblccc2bmdrv: done...\n\n" );

	return 0;
}
